/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ------------------------------------------------------------------------- $Id
 * : $ -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.jbi.JBIException;

import org.ow2.petals.bc.xquare.su.XQuareSUListener;
import org.ow2.petals.bc.xquare.util.DataSources;
import org.ow2.petals.component.framework.bc.AbstractBindingComponent;
import org.xquark.bridge.Mapping;
import org.xquark.bridge.XQBridge;

/**
 * <p>
 * XQuare BC JBI Component.
 * </p>
 * <p>
 * Registers a @ link XQuareSUManager} to handle SUs deployed onto the XQuare
 * BC.
 * </p>
 * 
 * @version $Rev: 8909 $ $Date: 2008-07-21 15:18:50 +0200 (lun., 21 juil. 2008)
 *          $
 * @since Petals 1.0
 * @author Adrien LOUIS - EBMWebSourcing
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 * @author Roland NAUDIN - EBM Websourcing
 */

/*
 * TODO return Fault for 'business' exception(IN document malformed), not ERROR
 */

/*
 * TODO use Transaction if defined in the Echange
 */

/*
 * TODO dabase notifications (Consumes) to be refactored (code not used)
 */

/*
 * TODO bridges and mapping xhould be stored and managed in a BridgeManager
 */

/*
 * FIXME pb with bridge.getMapping & bridge.getXMLConnection, so use 2 bridges
 */

public class XQuareBC extends AbstractBindingComponent {

    private DataSources dataSources;

    private Map<String, XQBridge> providerXQBridgesForMapping;

    private Map<String, XQBridge> providerXQBridgesForQuery;

    private Map<XQBridge, Mapping> xQBridgeMappings;

    /**
     * @deprecated not used yet
     */
    private Map<String, XQBridge> consumerXQBridges;

    /**
     * initialize the component
     * 
     * @throws JBIException
     *             impossible to initialize the component
     */
    @Override
    protected final void doInit() throws javax.jbi.JBIException {
        super.doInit();

        dataSources = new DataSources(getLogger());
        providerXQBridgesForMapping = Collections.synchronizedMap(new HashMap<String, XQBridge>());
        providerXQBridgesForQuery = Collections.synchronizedMap(new HashMap<String, XQBridge>());
        consumerXQBridges = Collections.synchronizedMap(new HashMap<String, XQBridge>());
        xQBridgeMappings = Collections.synchronizedMap(new HashMap<XQBridge, Mapping>());

        // set service unit listener
        final XQuareSUListener suListener = new XQuareSUListener(getLogger(), this);
        this.setServiceUnitListener(suListener);
    }

    public DataSources getDataSources() {
        return dataSources;
    }

    public Map<String, XQBridge> getProviderXQBridgesForMapping() {
        return providerXQBridgesForMapping;
    }

    public Map<String, XQBridge> getProviderXQBridgesForQuery() {
        return providerXQBridgesForQuery;
    }

    /**
     * @deprecated not used yet
     * @return
     */
    public Map<String, XQBridge> getConsumerXQBridges() {
        return consumerXQBridges;
    }

    public Map<XQBridge, Mapping> getXQBridgeMappings() {
        return xQBridgeMappings;
    }

}
