/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.consume;

import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_PREVIOUS_DELIMITER;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_UPDATE_DELIMITERS;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.StringTokenizer;
import java.util.logging.Logger;

import org.ow2.petals.bc.xquare.XQuareBCException;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.xquark.bridge.XQBridge;

/**
 * @version $Rev: $ $Date: $
 * @author alouis
 * @author Guillaume Decarnin - Open Wide
 *
 */
public class DatabaseDelimiterNewDataStrategy extends DelimiterNewDataStrategy {
    
    public void init(XQuareBCExternalListener processor, XQBridge xqBridge, ConfigurationExtensions extensions, Logger logger) throws XQuareBCException {
        super.init(processor, xqBridge, extensions, logger);

        // init from props :
        this.previousDataDelimiterSelect = extensions.get(XQUARE_DATALISTENER_PREVIOUS_DELIMITER);
        this.updateDelimiters = extensions.get(XQUARE_DATALISTENER_UPDATE_DELIMITERS);
        
        // initiating new data delimiter update statements :
        StringTokenizer stok = new StringTokenizer(updateDelimiters, ";");
        // splitting it to execute multiple updates if any
        // ex. for delimiter = max id : 1. cleaning the new data detection
        // table, 2. filling it with the new max id
        while (stok.hasMoreTokens()) {
            String newDataDelimiterUpdate = stok.nextToken().trim();
            newDataDelimiterUpdates.add(newDataDelimiterUpdate);
        }
    }

    public String getOldDataDelimiter() throws Exception {
        Connection connection = null;
    	Statement newDataStmt = null;
        try {
        	connection = dataSource.getConnection();
			newDataStmt = connection.createStatement();
			newDataStmt.execute(this.previousDataDelimiterSelect);
			ResultSet newDataRs = newDataStmt.getResultSet();
			if (newDataRs.next()) {
			    return newDataRs.getString(1);
			}
		} catch (Exception e) {
			throw e;
		} finally {
			if (newDataStmt != null) {
				newDataStmt.close();
			}
        	if (connection != null) {
                connection.close();
            }
		}
        return null;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.bc.xquarebc.listeners.NewDataStrategy#updateDataKnown()
     */
    public void updateDataKnown() throws Exception {
        // building SQL for updating the data delimiter to the new value
        // splitting it to execute multiple updates if any
        // ex. for delimiter = max id : 1. cleaning the new data detection
        // table, 2. filling it with the new max id
        for (String newDataDelimiterUpdate : newDataDelimiterUpdates) {
            newDataDelimiterUpdate = replaceDelimiterTokens(newDataDelimiterUpdate);
            Connection connection = null;
            Statement updateStatement = null;
            try {
            	connection = dataSource.getConnection();
				updateStatement = connection.createStatement();
				updateStatement.executeUpdate(newDataDelimiterUpdate);
				updateStatement.close();
			} catch (Exception e) {
				throw e;
			} finally {
	            if (updateStatement != null) {
	                try {
	                	updateStatement.close();
	                } catch (Exception e) {
	                    e.printStackTrace();
	                }
	            }
	        	if (connection != null) {
	                try {
	                	connection.close();
	                } catch (Exception e) {
	                	e.printStackTrace();
	                }
	            }
	        }
        }
    }
}
