/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.consume;

import java.util.Timer;
import java.util.TimerTask;

/**
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 * @author Roland NAUDIN - EBM Websourcing
 *
 */
public class DatabaseQueryPoller extends TimerTask {
    
    /**
     * The timer used for task scheduling
     */
    protected Timer timer;

    /**
     * The polling period
     */
    protected long period;

    /**
     * The first task execution delay
     */
    protected long delay;
    
    private XQuareBCExternalListener xQuareBCExternalListener;
    

    /**
     * Default constructor.
     * 
     * @param timer
     * @param period
     * @param delay
     * @param quareBCExternalListener
     */
    public DatabaseQueryPoller(long period, long delay,
            XQuareBCExternalListener quareBCExternalListener) {
        super();
        this.timer = new Timer();
        this.period = period;
        this.delay = delay;
        xQuareBCExternalListener = quareBCExternalListener;
    }
    
    /**
     * Start the task processing
     *
     */
    public void startProcessing() {
        timer.schedule(this, delay, period);
    }

    /**
     * Stop the task processing
     *
     */
    public void stopProcessing() {
        timer.cancel();
    }

    /* (non-Javadoc)
     * @see java.util.TimerTask#run()
     */
    @Override
    public void run() {
        // Handle the dataTable
        try {
            xQuareBCExternalListener.handleNewData();
        } catch (Exception e) {
            xQuareBCExternalListener.getLogger().severe(
                    "Error while handling new data " + "for target service " + xQuareBCExternalListener.service
                            + " with operation " + xQuareBCExternalListener.operation + " : " + e.getMessage());
        }

    }

}
