/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.consume;

import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_DELIMITER_FILENAME;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_DELIMITER_FOLDER;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DELIMITER_KEY;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Properties;
import java.util.logging.Logger;

import org.ow2.petals.bc.xquare.XQuareBCException;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.xquark.bridge.XQBridge;

/**
 * @version $Rev: $ $Date: $
 * @author alouis
 * @author Guillaume Decarnin - Open Wide
 *
 */
public class FileDelimiterNewDataStrategy extends DelimiterNewDataStrategy {

    private File file;
    
    public void init(XQuareBCExternalListener processor, XQBridge xqBridge, ConfigurationExtensions extensions, Logger logger) throws XQuareBCException {
        super.init(processor, xqBridge, extensions, logger);
        
        // get folder
        String folderPath = extensions.get(XQUARE_DATALISTENER_DELIMITER_FOLDER);
        if (folderPath == null || folderPath.trim().length() == 0) {
        	// default value: {petals-folder}/xquare-bc-polling-files
        	folderPath = System.getProperty("user.dir") + "/../xquare-bc-polling-files";
        }

        // get filename
        String fileName = extensions.get(XQUARE_DATALISTENER_DELIMITER_FILENAME);
        if (fileName == null || fileName.trim().length() == 0) {
        	// default value: ServiceName.properties
        	fileName = processor.service.getLocalPart() + ".properties";
        }
        
        // init file
        File folder = new File(folderPath);
        if (!folder.exists()) {
        	folder.mkdirs();
        }
        
        file = new File(folder, fileName);
    }

    public String getOldDataDelimiter() throws Exception {
        if (!file.exists()) {
            return null;
        }
        
        Properties propFile = new Properties();
        
        InputStream inputStream = null;
		try {
			inputStream = new FileInputStream(file);
			propFile.load(inputStream);
		} catch (Exception e) {
			throw e;
		} finally {
			if (inputStream != null) {
				inputStream.close();
			}
		}
        
        return propFile.getProperty(XQUARE_DELIMITER_KEY);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.bc.xquarebc.listeners.NewDataStrategy#updateDataKnown()
     */
    public void updateDataKnown() throws Exception {
        // updating the data delimiter in the file to the new value
        Properties propFile = new Properties();
        propFile.setProperty(XQUARE_DELIMITER_KEY, newDataDelimiter);
        
        OutputStream outputStream = null;
		try {
			outputStream = new FileOutputStream(file);
	        propFile.store(outputStream, null);
		} catch (Exception e) {
			throw e;
		} finally {
			if (outputStream != null) {
				outputStream.close();
			}
		}
    }
}
