/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ------------------------------------------------------------------------- $Id
 * : $ -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.provide;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.naming.NamingException;
import javax.sql.DataSource;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.bc.xquare.XQuareBC;
import org.ow2.petals.bc.xquare.XQuareBCException;
import org.ow2.petals.bc.xquare.XQuareConstants;
import org.ow2.petals.bc.xquare.util.StoredQuery;
import org.ow2.petals.bc.xquare.util.XQuareUtils;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xquark.bridge.Mapping;
import org.xquark.bridge.XQBridge;
import org.xquark.xml.xdbc.PreparedXMLStatement;
import org.xquark.xml.xdbc.XMLConnection;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLResultSet;
import org.xquark.xml.xdbc.XMLStatement;

/**
 * <p>
 * This processor builds upon the XQuare Bridge library (see web site, tutorial
 * and reference documentation at http://xquare.objectweb.org/ ) in order to
 * allow users to interact with databases through its four operations : insert,
 * query, storedQuery and executeSql.
 * </p>
 * 
 * <h4>Operations detail</h4>
 * 
 * <p>
 * The "insert" operation lets the user provide as the content of an incoming
 * JBI message an XML document defining insertion or update of database records
 * according to the configured XQuare mapping (mapping XML file and insertion
 * XML schema file). Note that updating existing data or adding new records are
 * both possible. Look at the XQuare Bridge reference documentation for more
 * details. The content of the response is (if no fault because of an internal
 * exception) the single tag "<success/>".
 * </p>
 * 
 * <p>
 * The "query" operation lets the user provide as the content of an incoming JBI
 * message an XQuery wrapped within the "query" top tag. This XQuery must be
 * XQuare compliant and consistent with the configured database and mapping for
 * the targeted service. It should also be XML encoded (even if the parser tries
 * to be nice by formatting and including XML sub-nodes) - typically the results
 * formatting should be, since they comprise XML tags. The response is an XML
 * document whose top tag is "&lt;Results&gt;" and whose elements are specified
 * by the XQuery's results formatting part. Note that modules (akin to XML / SQL
 * views, defined in *.mod files provided in the service unit's deployment
 * directory) can be imported and used in such queries. Look at the XQuare
 * XQuery reference documentation for more details.
 * </p>
 * 
 * <p>
 * There is a way to specify the top tag in the query : by writing it down in
 * xquery (ex. &lt;query&gt;&lt;result&gt;{ for $i in...) and using a service
 * configured with the queryResults.wrapInRootTag at false.
 * </p>
 * 
 * <p>
 * The "storedQuery" operation lets the user provide as the content of an
 * incoming JBI message the id of the stored parameterized query to be executed
 * and the value of its parameters. Concretely it must consist in the
 * "storedQuery" top tag with the "id" attribute and zero or more sub-tags with
 * "name" and "value" attributes. Those "names" must be consistent with the
 * parameters defined within the XQuery code of the stored query (see examples),
 * since the "parameter" mechanism of XQuare / XQuery is used by the processor.
 * Otherwise it is similar to the "query" operation in its behaviour.
 * </p>
 * 
 * <h4>Configuration</h4>
 * 
 * <p>
 * The target database is configured through the SU extensions provided by a
 * given service unit and uses the following properties :
 * </p>
 * <ul>
 * <li>xquare.database.driver : the JDBC driver class name of the targeted
 * database. Required for accessing the database through JDBC. Ex.
 * com.mysql.jdbc.Driver</li>
 * <li>xquare.database.url : the JDBC URL of the targeted database. Required for
 * accessing the database through JDBC. Ex.
 * jdbc:mysql://localhost/petals_samples_xquare_auction</li>
 * <li>xquare.database.user : the user required by the JDBC connection. Required
 * for accessing the database through JDBC. Ex. petals_samples</li>
 * <li>xquare.database.password : the password required by the JDBC connection.
 * Required for accessing the database through JDBC. Ex. petals_samples</li>
 * <li>xquare.jndidatasource.name : if provided, is used as the preferred way to
 * access the target database. Note that the given JNDI name must exist in the
 * Petals context and target a J2EE Datasource Ex. jdbc/myDS</li>
 * </ul>
 * 
 * <p>
 * Additional properties for configuring XQuare Bridge :
 * </p>
 * <ul>
 * <li>xquare.database.mapping : the name of the XQuare mapping file to be found
 * in the service unit deployment directory Required. Ex. auction.map</li>
 * <li>xquare.queryResults.wrapInRootTag : whether to always wrap the query
 * results in a top "&lt;Results&gt;" tag. If set to false, single results won't
 * be wrapped.</li>
 * <li>xquare.xmlconnection.baseuri : The base URI where to find modules (.mod
 * files) to import if such are specified in user or stored queries. Optional,
 * defaults to the service unit's deployment directory</li>
 * <li>(not a property per su) The schema definition of insertion XML documents
 * must be provided in the service unit's deployment directory, with the same
 * prefix than the corresponding mapping file (ex. auction.xsd)</li>
 * </ul>
 * 
 * <p>
 * Configuring stored queries :<br/> Stored queries must be defined in a tag
 * "&lt;storedQuery id="x"&gt;". The query id is a String and is its XQuery
 * definition. Note that parameters are implicitly defined using the
 * corresponding XQuare / XQuery concept. See examples.
 * </p>
 * 
 * @version $Rev: 8993 $ $Date: 2008-07-16 15:57:24 +0200 (mer., 16 juil. 2008)
 *  $
 * @since Petals 1.0
 * @author alouis
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 * 
 */

/*
 * TODO move constants
 */

/*
 * TODO need more refactoring
 */

public class XQuareBCJBIListener extends AbstractJBIListener {

    private static final String NS = "http://petals.ow2.org/components/xquare";

    private static final String INSERT_OPERATION = "insert";

    private static final String UPDATE_OPERATION = "update";

    private static final String QUERY_OPERATION = "query";

    private static final String EXECUTE_SQL_OPERATION = "sql";

    private static final String SUCCESS_TAG = "success";

    private static final String QUERY_TAG = "query";

    private static final String SQL_TAG = "sql";

    private static final String STORED_QUERY_OPERATION = "storedQuery";

    private static final String STORED_QUERY_TAG = "storedQuery";

    private static final String STORED_QUERY_ID_ATTR = "id";

    private static final String STORED_QUERY_PARAM_TAG = "param";

    private static final String STORED_QUERY_PARAM_NAME_ATTR = "name";

    private static String SUCCESS_RESPONSE;

    public XQuareBCJBIListener() {
        SUCCESS_RESPONSE = "<xquare:" + SUCCESS_TAG + " xmlns:xquare=\"" + NS + "\"/>";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.listener.AbstractJBIListener#onJBIMessage
     * (org.ow2.petals.component.framework.util.Exchange)
     */
    @Override
    protected void onJBIMessage(Exchange exchange) {
        try {
            String operation = exchange.getOperation().getLocalPart();
            getLogger()
                    .finest(
                            exchange.getExchangeId() + " - Exchange received with operation : "
                                    + operation);

            if (INSERT_OPERATION.equals(operation) || UPDATE_OPERATION.equals(operation)) {
                getLogger().finest(exchange.getExchangeId() + " - Start the insert operation");
                this.doInsert(exchange);
            } else if (QUERY_OPERATION.equals(operation)) {
                this.doQuery(exchange);
            } else if (STORED_QUERY_OPERATION.equals(operation)) {
                doStoredQuery(exchange);
            } else if (EXECUTE_SQL_OPERATION.equals(operation)) {
                doExecuteSql(exchange);
            } else {
                throw new XQuareBCException("Unknown operation " + operation);
            }
        } catch (MessagingException e) {
            getLogger().log(Level.SEVERE, e.getMessage(), e);
            exchange.setError(e);
        } catch (Exception e) {
            getLogger().log(Level.SEVERE, e.getMessage(), e);
            exchange.setError(new MessagingException("Failed to process the exchange : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage()));
        }
    }

    /**
     * Implementation of the "insert" operation.
     * 
     * @param exchange
     *      service unit extensions
     * @throws XQuareBCException
     * @throws MessagingException
     * @throws PEtALSCDKException
     * @throws SAXException
     * @throws XMLDBCException
     */
    protected void doInsert(Exchange exchange) throws XQuareBCException, PEtALSCDKException,
            MessagingException, XMLDBCException, SAXException {
        getLogger().finest(exchange.getExchangeId() + " - Start insert operation");
        // Loading the XQuare bridge of the specified service unit
        String ref = XQuareUtils.generateKey(provides);
        getLogger().finest(exchange.getExchangeId() + " - Retrieve the XQBridge mapper for " + ref);
        XQBridge bridge = ((XQuareBC) this.getComponent()).getProviderXQBridgesForMapping()
                .get(ref);

        // FIXME: Use the document directly when Xquare would be DOM3
        // compatible
        getLogger().finest(exchange.getExchangeId() + " - Insert document");
        Mapping mapping = ((XQuareBC) this.getComponent()).getXQBridgeMappings().get(bridge);
        mapping.getMapper().insertDocument(exchange.getInMessageContentAsString());

        // Create out "success" message
        if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {
            StreamSource source = new StreamSource(new ByteArrayInputStream(SUCCESS_RESPONSE
                    .getBytes()));
            exchange.setOutMessageContent(source);
        }
        getLogger().finest(exchange.getExchangeId() + " - End of the insert operation");
    }

    /**
     * Implementation of the "query" operation.
     * 
     * @param exchange
     *      must be an XQuery consistent with the configured database and
     *      mapping of the specified service unit. It should be XML-encoded and
     *      wrapped in a "query" top tag (even if the parser tries to be nice by
     *      formatting and including XML subnodes).
     * @throws MessagingException
     * @throws PEtALSCDKException
     * @throws XQuareBCException
     * @throws XMLDBCException
     */
    protected void doQuery(Exchange exchange) throws MessagingException, PEtALSCDKException,
            XQuareBCException, XMLDBCException {
        getLogger().finest(exchange.getExchangeId() + " - Start the query operation");
        // refuse MEP with no OUT message
        if (exchange.isInOnlyPattern() || exchange.isRobustInOnlyPattern()) {
            throw new MessagingException("The MEP '" + exchange.getExchangePattern()
                    + "' is not supported for operation '" + QUERY_OPERATION + "'");
        }

        // TODO : temporary trick for a urgent customer release, must be removed
        // after
        String queryString = exchange.getInMessageContentAsString();
        queryString = removeXMLTopElement(queryString, QUERY_TAG);

        // Document queryDocument = exchange.getInMessageContentAsDocument();
        // String queryString = XQuareUtils.getRootElementContent(queryDocument,
        // QUERY_TAG);

        XMLConnection xc = null;
        try {
            // Loading the XQuare bridge of the specified service unit
            String ref = XQuareUtils.generateKey(provides);
            getLogger().finest(exchange.getExchangeId() + " - Retrieve the XQBridge for " + ref);
            XQBridge bridge = ((XQuareBC) this.getComponent()).getProviderXQBridgesForQuery().get(
                    ref);

            // Getting the XML connection
            getLogger().finest(exchange.getExchangeId() + " - Get an XML connection");
            xc = bridge.getXMLConnection();

            // Creating the corresponding XML Statement
            getLogger().finest(exchange.getExchangeId() + " - Prepare statement");
            XMLStatement xs = xc.createStatement();

            // set baseURI
            String baseURI = this.getExtensions()
                    .get(XQuareConstants.XQUARE_XML_STATEMENT_BASE_URI);
            if (baseURI == null) {
                baseURI = "file://" + this.getInstallRoot();
            }
            xs.setBaseURI(baseURI);

            // executing the query
            getLogger().finest(exchange.getExchangeId() + " - Execute the xquery");
            XMLResultSet xrs = xs.executeQuery(queryString);

            getLogger().finest(
                    exchange.getExchangeId() + " - Convert the result to an XML document");
            Document resXmlDocument = XQuareUtils.convertToDocument(xrs);

            exchange.setOutMessageContent(resXmlDocument);
            getLogger().finest(exchange.getExchangeId() + " - End of the query operation");
        } finally {
            try {
                // properly closing the XML connection and the resources it
                // relies on
                getLogger().finest(exchange.getExchangeId() + " - Close the XML connection");
                xc.close();
            } catch (XMLDBCException e1) {
                this.getLogger().log(Level.WARNING, "Error closing XML connection");
            }
        }
    }

    /**
     * Implementation of the "storedQuery" operation.
     * 
     * @param exchange
     * @return a Source corresponding of the formatted results of the stored
     *  XQuery, wrapped within the "Results" top tag.
     * @throws XQuareBCException
     * @throws MessagingException
     */
    protected void doStoredQuery(Exchange exchange) throws XQuareBCException, MessagingException {

        getLogger().finest(exchange.getExchangeId() + " - Start the storedquery operation");
        // refuse MEP with no OUT message
        if (exchange.isInOnlyPattern() || exchange.isRobustInOnlyPattern()) {
            throw new MessagingException("The MEP '" + exchange.getExchangePattern()
                    + "' is not supported for operation '" + STORED_QUERY_OPERATION + "'");
        }

        Document contentDocument = exchange.getInMessageContentAsDocument();
        Element queryElement = contentDocument.getDocumentElement();
        if (!STORED_QUERY_TAG.equals(queryElement.getLocalName())) {
            throw new XQuareBCException("XQuery statement(s) must have a root element named '"
                    + STORED_QUERY_TAG + "'");
        }
        Node contentNode = queryElement.getFirstChild();

        // load from extensions the stored queries and their query
        // template
        StoredQuery storedQuery = getStoredQuery(contentNode);
        getLogger()
                .finest(
                        exchange.getExchangeId() + " - Stored query to use : "
                                + storedQuery.getStatement());
        getLogger().finest(
                exchange.getExchangeId() + " - Stored query parameters : "
                        + storedQuery.getParameterMap());

        // TODO: validate the parameters??

        XMLConnection xc = null;
        Document resXmlDocument = null;

        // Loading the XQuare bridge of the specified service unit
        String ref = XQuareUtils.generateKey(provides);
        getLogger().finest(exchange.getExchangeId() + " - Retrieve the XQBridge mapper for " + ref);
        XQBridge bridge = ((XQuareBC) this.getComponent()).getProviderXQBridgesForQuery().get(ref);

        try {
            getLogger().finest(exchange.getExchangeId() + " - Get an XML connection");
            xc = bridge.getXMLConnection();
        } catch (XMLDBCException e) {
            this.getLogger()
                    .log(Level.SEVERE, exchange.getExchangeId() + " - " + e.getMessage(), e);
            throw new XQuareBCException("XQuare XML DB error when getting XML connection", e);
        }
        try {
            File suRootFile = new File(this.getInstallRoot());
            String baseUri = this.getExtensions()
                    .get(XQuareConstants.XQUARE_XML_STATEMENT_BASE_URI);
            if (baseUri == null || baseUri.trim().length() == 0) {
                baseUri = suRootFile.toURI().toString();
            }

            xc.setBaseURI(baseUri);
            // loading the corresponding xquare prepared statement
            getLogger().finest(
                    exchange.getExchangeId() + " - Prepare statement from the stored xquery");
            PreparedXMLStatement pxs = xc.prepareStatement(storedQuery.getStatement());

            // filling the query template with parameters
            for (String key : storedQuery.getParameterMap().keySet()) {
                String value = storedQuery.getParameterMap().get(key);
                pxs.setString(key, value);
            }

            // executing the query
            getLogger().finest(exchange.getExchangeId() + " - Execute the xquery");
            XMLResultSet xrs = pxs.executeQuery();
            // NB. here could be done line number limit, pagination...

            // formating the results
            getLogger().finest(
                    exchange.getExchangeId() + " - Convert the result to an XML document");
            resXmlDocument = XQuareUtils.convertToDocument(xrs);

            // Create out message with the returned XML results
            // TODO: XQuark fails silently ?! Check insert result
            // in db ??
            exchange.setOutMessageContent(resXmlDocument);
            getLogger().finest(exchange.getExchangeId() + " - End of the storedquery operation");

        } catch (Exception e) {
            this.getLogger()
                    .log(Level.SEVERE, exchange.getExchangeId() + " - " + e.getMessage(), e);
            throw new XQuareBCException("XQuare XML DB error when trying to query", e);
        } finally {
            try {
                // properly closing the XML connection and the resources it
                // relies on
                getLogger().finest(exchange.getExchangeId() + " - Close the XML connection");
                xc.close();
            } catch (XMLDBCException e1) {
                this.getLogger().log(Level.WARNING, "Error closing XML connection");
            }
        }
    }

    /**
     * Builds the stored query model.
     * 
     * @param contentNode
     *      the node to build it from
     * @param extensions
     *      the extensions to check the provided query id against
     * @return a filled StoredQuery model
     * @throws XQuareBCException
     *      if any error
     */
    protected StoredQuery getStoredQuery(Node contentNode) throws XQuareBCException {
        StoredQuery storedQuery = new StoredQuery();

        // get query id
        String storedQueryId = XMLUtil.getAttributeValue(contentNode, STORED_QUERY_ID_ATTR);
        if (storedQueryId == null || storedQueryId.length() == 0) {
            throw new XQuareBCException("Found stored query element without an id !");
        }
        storedQuery.setId(storedQueryId);

        String storedQueryStatement = null;
        // search the storedQuery in the SU extensions.
        // if there is only one stored query, it is referenced as "stored-query"
        // if there is more than one, they are references as "stored-query1"...
        storedQueryStatement = this.getExtensions().get(
                XQuareConstants.XQUARE_STORED_QUERY_ID + storedQueryId);
        if ("1".equals(storedQueryId) && storedQueryStatement == null) {
            storedQueryStatement = this.getExtensions().get(XQuareConstants.XQUARE_STORED_QUERY_ID);
        }

        // get query statements
        if (storedQueryStatement == null || storedQueryStatement.trim().length() == 0) {
            throw new XQuareBCException("Stored query with id " + storedQueryId
                    + " has no query statement !");
        }
        storedQuery.setStatement(storedQueryStatement);

        // get params)
        HashMap<String, String> storedQueryParamMap = new HashMap<String, String>();
        Node currentParamNode = XMLUtil.getFirstChild(contentNode);
        while (currentParamNode != null) {
            if (STORED_QUERY_PARAM_TAG.equalsIgnoreCase(currentParamNode.getNodeName())) {
                String paramName = XMLUtil.getAttributeValue(currentParamNode,
                        STORED_QUERY_PARAM_NAME_ATTR);
                if (paramName == null || paramName.length() == 0) {
                    throw new XQuareBCException("Found stored query parameter without a name "
                            + "for stored query with id " + storedQueryId);
                }
                String paramValue = XMLUtil.getTextContent(currentParamNode);
                storedQueryParamMap.put(paramName, paramValue);
            }

            currentParamNode = XMLUtil.getNextSibling(currentParamNode);
        }
        storedQuery.setParameterMap(storedQueryParamMap);

        return storedQuery;
    }

    /**
     * Implementation of the "executeSql" operation. Use a standard JDBC
     * mechanism (xquare not used)
     * 
     * @param exchange
     *      service unit extensions
     * @throws MessagingException
     * @throws NamingException
     * @throws SQLException
     * @throws PEtALSCDKException
     * @throws ParserConfigurationException
     */
    protected void doExecuteSql(Exchange exchange) throws MessagingException, PEtALSCDKException,
            SQLException, NamingException, ParserConfigurationException, XQuareBCException {

        getLogger().finest(exchange.getExchangeId() + " - Start the sql operation");

        Connection connection = null;
        Statement statement = null;

        Document queryDocument = exchange.getInMessageContentAsDocument();
        String sqlString = XQuareUtils.getRootElementContent(queryDocument, SQL_TAG);

        String[] sqls;
        if (sqlString.contains(";")) {
            sqls = sqlString.split(";"); // TODO bug with sql like
        } else {
            sqls = new String[] { sqlString };
        }
        //' my;string'
        DataSource ds = ((XQuareBC) this.getComponent()).getDataSources().getDataSource(
                getExtensions(), null, false);
        connection = ds.getConnection();
        statement = connection.createStatement();
        connection.setAutoCommit(false);

        try {
            for (int i = 0; i < sqls.length - 1; i++) {
                String query = sqls[i];
                if (query == null || query.trim().length() == 0) {
                    continue;
                }
                statement.execute(query);
            }
            // the last statement can be a select
            String lastStatement = sqls[sqls.length - 1];
            ResultSet resultSet = null;
            if (lastStatement != null && lastStatement.trim().length() > 0) {
                if (lastStatement.trim().startsWith("select") || lastStatement.trim().startsWith("SELECT")) {
                    resultSet = statement.executeQuery(lastStatement);
                } else {
                    statement.execute(lastStatement);
                }
            }

            Document resultDocument = null;
            if (resultSet != null) {
                resultDocument = XQuareUtils.convertToRawDocument(resultSet);

            }

            connection.commit();
            
            if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {
                if (resultDocument != null) {
                    exchange.setOutMessageContent(resultDocument);
                } else {
                    StreamSource result = new StreamSource(new ByteArrayInputStream(
                            SUCCESS_RESPONSE.getBytes()));
                    exchange.setOutMessageContent(result);
                }
            }
            getLogger().finest(exchange.getExchangeId() + " - End of the sql operation");
        } finally {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    getLogger().warning("Error while closing SQL statement");
                }
            }
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    getLogger().warning("Error while closing SQL connection");
                }
            }
        }
    }

    /**
     * Remove the xml top element of the given xml document. String methods
     * usage, may not be good for large documents
     * 
     * @param xmlDocument
     * @param topElementName
     *      the top element, namespace is not needed
     * @return
     */
    public static final String removeXMLTopElement(String xmlDocument, String topElementName) {
        String result = null;
        if (xmlDocument != null && topElementName != null) {
            // 1 find the beginning of the sub document (assume that namespace
            // can appear)
            int topElementStartIndex = xmlDocument.indexOf(topElementName);
            if (topElementStartIndex >= 0) {
                String next = xmlDocument.substring(topElementStartIndex, xmlDocument.length());
                int startIndex = next.indexOf('>');
                // 2 check the top element
                if (startIndex >= 0) {
                    next = next.substring(startIndex + 1, next.length());
                    // 3 find the end of the sub document(assume that namespace
                    // can appear)
                    int topElementEndIndex = next.lastIndexOf(topElementName);
                    if (topElementEndIndex > 0) {
                        next = next.substring(0, topElementEndIndex);
                        int endIndex = next.lastIndexOf('<');
                        if (endIndex > 0) {
                            result = next.substring(0, endIndex);
                        }
                    }
                }
            }
        }
        return result;
    }
}
